/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

/*******************************************************************************
* WRITEPRETTYBASE output basecalls in prettybase format
*   WRITEPRETTYBASE(startpos, ref_name, hap1, hap2, ref, filename) will 
*   output basecalls in hap1 and hap2 in prettybase format in file
*   filename.
*   
*   Prettybase format is (tab-delimited):
*       $ref_$pos	$ref_name	$Hap1	$Hap2	$type
*   where $pos is the position, $ref_name is the reference name, $Hap1 is
*   (are) the base(s) of haplome 1 at that position, and similar for $Hap2.
*   $type is the type of call: R for no change from reference, H is for a
*   heterozygous snp, and A is for a Homozygous SNP.
*
*   N.B. on interpretation: gaps in reference are illustrated by multi-base
*   hap1 and hap2 entries. So an entry of the like:
*       ref_99812   HDF_3   ACCG	ACCG    A
*   means that as after position 99812 there are insertions 'CCG' in both
*   haplomes.
*
*   Parameters:
*   starpos - the starting position of this region/amplicon. 
*   ref_name - the reference name 
*   don_name - the donor name as described above
*   hap1 - the first haplome base calls, in ns or ls
*   hap2 - the first haplome base calls, (in ns or ls, but same format as 
*       hap1)
*   ref - the original reference, (in ns or ls, but same format as hap1)
*   filename - the filename to output to
*   
*   See also READPRETTYBASE
*
*   Author: Adrian V. Dalca
*******************************************************************************/
void writePrettybase(
    VA_UINT32    startpos, 
    const char * ref_name, 
    VA_VecU8     hap1, 
    VA_VecU8     hap2, 
    VA_VecU8     ref, 
    VA_MatDouble maxvals_2)
{
    int h1t = TRUE;
    int h2t = TRUE;
    int h3t = TRUE;
    VA_VecU8 don1;
    VA_VecU8 don2;
    VA_VecU8 ref2;
    int i;
    int pos;
    char * oline = "%s_%d\t%s\t%s\t%s\t%c";
    char * oline2 = "%s_%d\t%s\t%c\t%c\t%c";




    /* transform the data to ls format (a,c,g,t,-,n) */
    for (i = 0; i < hap1.size; i++) {
        if (isalnum(hap1.data[i]) == 0)
            h1t = FALSE;
        if (isalnum(hap2.data[i]) == 0)
            h2t = FALSE;
        if (isalnum(ref.data[i]) == 0)
            h3t = FALSE;
    }
        
    VA_ASSERT(h1t == h2t && h1t == h3t);
    if (!h1t) {
        don1 = nuc2nuc(hap1);
        don2 = nuc2nuc(hap2);
        ref2 = nuc2nuc(ref);
    } else {
        don1 = va_allocVecU8(hap1.size);
        don2 = va_allocVecU8(hap1.size);
        ref2 = va_allocVecU8(ref.size);

        don1 = hap1;
        don2 = hap2;
        memcpy(don1.data, hap1.data, sizeof(VA_UINT8) * don1.size);
        memcpy(don2.data, hap2.data, sizeof(VA_UINT8) * don2.size);
        memcpy(ref2.data, ref.data, sizeof(VA_UINT8) * ref2.size);
    }
    


    
    /* go through each position */
    i = 0;
    pos = 1;
    while (i < ref2.size) {
        char rha;

        /* if ref does not have a gap */
        if (ref2.data[i] != '-') {
            if ((don1.data[i] == don2.data[i]) && 
                (don1.data[i] == 'n')) {
                rha = 'R';
            } else if ((don1.data[i] == don2.data[i]) && 
                     (don1.data[i] != ref2.data[i])) {
                rha = 'A';
            } else if (don1.data[i] != don2.data[i]) {
                /* neither should be 'n'. TODO: should we check for this? */
                rha = 'H'; 
	    } else if ((don1.data[i] == don2.data[i]) && 
                     (don1.data[i] == ref2.data[i])) {
                rha = 'R';
            } else {
                /* should not get here */
                VA_ASSERT(FALSE);
            }
            
            /* if the next ref charactes is NOT a gap, write to file */
            if ((i == ref2.size - 1) || (ref2.data[i + 1] != '-')) {
		    if (predict_start==0 || (pos>=(predict_start-hmm_start+1) && pos <=(predict_end-hmm_start+1))) {
		    if (rha!='R' || report_all==1) {
			fprintf(output_file, oline2, ref_name, startpos + pos - 1 + (predict_start>0 ? hmm_start -1 : 0), donor_name, 
				toupper(don1.data[i]), toupper(don2.data[i]), rha);
			if (viterbi==1) {
				fprintf(output_file,"\n");
			} else {
				fprintf(output_file,"\t%4.2f\n",maxvals_2.data[i]);
			}
		    }
		    }
		    pos = pos + 1;
	    }
            
            i = i + 1;
        } else if (i>0) {
            /* there is a gap in the reference */
            int gaplen;
            char * d1;
            char * d2;
            int j;
            int k;
            char rhamode;
            
            /* make sure we're not all gaps... */
            if ((ref2.data[i] == '-') && 
                (don1.data[i] == '-') && 
                (don2.data[i] == '-') &&
		(ref2.data[i-1] == '-')) {
                fprintf(stderr, "writePrettybase error:\n\t"
                    "ref, hap1 & hap2 all have gaps at position%d\n", i + 1);
                
                i = i + 1;
                continue;
            }
            
            /* get the length of the gap. */
            gaplen = 1;
            j = i + 1;
            while (j<ref2.size && ref2.data[j] == '-') {
                gaplen++;
                j++;
            }
	    int calling_reference = (ref2.data[i-1]==don1.data[i-1] && ref2.data[i-1]==don2.data[i-1]) || (don1.data[i-1]=='n' && don2.data[i-1]=='n'); // case of      A----/A---- when ref is A
	    for (k=i+1; k<ref2.size && ( (don1.data[k]=='-' && don2.data[k]=='-') || (don1.data[k]=='n' && don2.data[k]=='n')) && ref2.data[k]=='-'; k++);
	    calling_reference = calling_reference && (k==j);
            
            /* check weather the indel is het/hom. */
            int homozygous = TRUE;
            for (k = i - 1; k <= i + gaplen - 1; k++) {
                if (hap1.data[k - 1] != hap2.data[k - 1]) {
                    homozygous = FALSE;
                    break;
                }
            }
            rhamode = calling_reference ? 'R' : ((homozygous) ? 'A' : 'H');
	    //fprintf(output_file, "  cr: %d, hm: %d, k %d, j %d, %c %c %c\n",calling_reference,homozygous,k,j,ref2.data[i],don1.data[i],don2.data[i]);

            
            d1 = (char *)malloc(gaplen + 3);
            VA_ASSERT(d1 != NULL);
            d2 = (char *)malloc(gaplen + 3);
            VA_ASSERT(d2 != NULL);

            for (k = i - 1; k <= i + gaplen - 1; k++) { 
                d1[k - i + 1] = toupper(don1.data[k]);
                d2[k - i + 1] = toupper(don2.data[k]);
            }
	    int printed_gap_len=gaplen;
	    while (printed_gap_len>0 && d1[printed_gap_len]=='-' && d2[printed_gap_len]=='-'){
	 	printed_gap_len--;	
	    }
            d1[printed_gap_len + 1] = '\0';
            d2[printed_gap_len + 1] = '\0';
	    if (pos==1) {
		d1[printed_gap_len+1]=toupper(don1.data[i+gaplen]);
		d1[printed_gap_len+2]='\0';
		d2[printed_gap_len+1]=toupper(don2.data[i+gaplen]);
		d2[printed_gap_len+2]='\0';
		i++;
	    }
	    if (calling_reference) {
		d1[1]='\0';
		d2[1]='\0';
	    }
	    if (predict_start==0 || (pos>=(predict_start-hmm_start+1) && pos <=(predict_end-hmm_start+1))) {
		    if (rhamode!='R' || report_all==1) {
		    fprintf(output_file, oline, ref_name, startpos + pos - 1 + (predict_start>0 ? hmm_start -1 : 0), 
			    donor_name, d1, d2, rhamode);
		if (viterbi==1) {
			fprintf(output_file,"\n");
		} else {
			fprintf(output_file,"\t%4.2f\n",maxvals_2.data[i]);
		}
		    }
	    }
		
            pos = pos + 1;

            
            i = i + gaplen;
            free(d1);
            free(d2);
        } else {
	    i++;
	}
    }    
    
    free(don1.data);
    free(don2.data);
    free(ref2.data);

    /*
     * TODO test: there are 8 cases that should be tested in more detail:
     * 4 with no gap in ref - these should be easy
     * 4 more with a gap in ref. (hap1 x hap2 gap on/off). these can be harder
     *   and of multiple lengths...
     */
}
