/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include "va_log.h"
#include "va_assert.h"
#include "varid.h"
#include "va_time.h"

#define READSIZE_CUTOFF (33)
#define NORMODDS_CUTOFF (0)
#define PGENOME_CUTOFF  (0)
#define MAPQ_CUTOFF		(-1)

#define ALLELE_TOP      (2)
#define ALLELE_BOT      (1)
#define ALLELE_OK       (4)
#define ALLELE_NONE     (3)

static VA_UINT8 destroy(
    CData_t     cdata,
    VA_Vector   reads,
    VA_VecI32   readstarts,
    Covmaps_t   covmap,
    VA_UINT32   f1,
    VA_UINT32   f2,
    VA_VecU32   diColor1,
    VA_VecU32   diColor2,
    VA_UINT32   letter1,
    VA_UINT32   letter2);

static  VA_VecU8  refcol;
/******************************************************************************
*   creates a list of states that need to be modified after the VARiD
*   algorithm has been run.
*       [ch_list, tot, top2] = destroy_state(cmat, reads, maxreadsize, 
*           readstarts, covmap, thr_vec, gnr, snps) creates a list of
*           states that need to be modified after the VARiD algorithm has
*           been run. 
*
*   Parameters:
*   cmat - the color matrix for the two alleles
*   reads - the set of reads 
*   maxreadsize - the maximum read size
*   readstarts - a vector of all the read starting positions
*   covmap - the covmap as computed in varid_covmap.m
*   thr-vec the vector of thresholds to use when trehsolding support
*       needed for destorying states
*   gnr - the group number (used for debugging)
*   snps - optional (used for debugging) - a boolean vector of the real snp
*       calls
*
*   Returns:
*   ch_list - size_thr x gen_len indicating destroyed positions at a
*       specific treshold
*   tot - a 2 x size_thr vector re-iterating the threshold and the amount
*       of total support
*   top2  - a struct array of size of size_thr in with fields:
*       top - a 2 x #color_disagreements vector maintaing the best alleles
*       bottom - a 2 x #color_disagreements vector maintaing the
*           second-best alleles (not sure here...)
*       pos - the position of the deletion.
*
*   New methodology written 10/4/2009, 12:40AM
*
*   Author: Adrian V. Dalca
*
*
*******************************************************************************/
void    destroy_state(
    CData_t        cdata,
    VA_MatU8       cmat, 
    VA_Vector      reads,
    VA_VecI32      readstarts, 
    Covmaps_t      covmap, 
    VA_VecDouble   thr_vec, 
    VA_MatU8       lmat,
    VA_MatU8     * ch_list,
    VA_MatDouble * tot, 
    VA_MatDouble * top2)
{
    struct timespec ts;
    struct timespec te;
    struct timespec tdiff;
    int rc;
    int i;
    int j;
    int t;
    VA_VecI32 f;
    VA_VecI32 z;
    VA_VecI32 pair_starts;
    VA_VecI32 pair_ends;
    VA_UINT32 f0;
    VA_UINT32 f1;
    VA_UINT32 f2;
    VA_UINT32 ok;
    VA_VecU32 diColor1;
    VA_VecU32 diColor2;
    VA_UINT32 letter1;
    VA_UINT32 letter2;
    VA_VecU8  refcol2;

    VA_ASSERT(ch_list != NULL);
    VA_ASSERT(tot != NULL);
    VA_ASSERT(top2 != NULL);

    memset(ch_list, 0, sizeof(ch_list));


    memset(&f, 0, sizeof(f));
    memset(&z, 0, sizeof(z));
    memset(&pair_starts, 0, sizeof(pair_starts));
    memset(&pair_ends, 0, sizeof(pair_ends));

    /* TODO: should add error rates and quality values to this
     * calculation
     * TODO: allow pairs that are DELAYED by gaps. i.e what if there's a snp right
     * before a gap...
     */

    
    refcol2 = nuc2col(cdata.ref);
    refcol = va_allocVecU8(refcol2.size + 1);

    refcol.data[0] = 0;
    for (i = 1; i < refcol.size; i++)
        refcol.data[i] = refcol2.data[i - 1]; 

    free(refcol2.data);

    memset(&top2, 0, sizeof(top2));
    memset(&tot, 0, sizeof(tot));
    
    /* prepare the SNP altering matrix */
    *ch_list = va_allocMatU8(cmat.size_x, thr_vec.size); 
    memset(ch_list->data, 0, 
           ch_list->size_x * ch_list->size_y * sizeof(VA_UINT8));
    if (no_heuristics) {
	return;
    }

    rc = va_GetTime(&ts);
    VA_ASSERT(rc == 0);

    fprintf(stderr, "\tStarting destroy_state\n");
    /* find the positions with different color calls */
    f = va_allocVecI32(cmat.size_x);
    f.size = 0;
    for (i = 0; i < cmat.size_x; i++) {
        if ((cmat.data[i] != cmat.data[cmat.size_x + i]) ||
           (cmat.data[i] != refcol.data[i]) ||
           (cmat.data[cmat.size_x + i] != refcol.data[i])) {
            f.data[f.size++] = i + 1;
        }
    }
    pair_starts.size=0;
    pair_starts.data=NULL;    
    pair_ends.size=0;
    pair_ends.data=NULL;
    if (f.size > 0) {
        /* find pairs */
        f = va_reallocVecI32(f, f.size);
        pair_starts = va_allocVecI32(f.size);
        pair_starts.size = 0;
        for (i = 0; i < f.size-1; i++)
            if (f.data[1 + i] - f.data[i] == 1)
                pair_starts.data[pair_starts.size++] = f.data[i];

        if (pair_starts.size > 0) {
            pair_starts = va_reallocVecI32(pair_starts, pair_starts.size);
            pair_ends = va_allocVecI32(pair_starts.size);
            for (i = 0; i < pair_ends.size; i++)
                pair_ends.data[i] = pair_starts.data[i] + 1;
        } else {
            free(pair_starts.data);
            pair_starts.data = NULL;
            free(pair_ends.data);
            pair_ends.data = NULL;
        }

    }
    VA_INT32 * f_orig = f.data;
    while (f.size > 0) {
        int toskip = FALSE;
	//check if it is a pair?
        for (i = 0; i < pair_starts.size; i++) {
            if (f.data[0] == pair_starts.data[i]) {
                toskip = TRUE;
                break;
            }
        }

        if (toskip) {
            f.size--;
            f.data = &f.data[1];
    
            if (f.size > 0) { 
                if (lmat.data[f.data[0] - 1] == 4 ||
                    lmat.data[lmat.size_x + f.data[0] - 1] == 4)  {
                    	f.size--;
                    	f.data = &f.data[1];
               }
            }

            continue;
        } 
    
        if (f.size == 1) {
            for (i = 0; i < ch_list->size_y; i++)
                ch_list->data[i * ch_list->size_x + f.data[0] - 1] = ALLELE_NONE;
            f.size--;
            f.data = &f.data[1];
            break;
        }
        
        /* otherwise. */
        f1 = f.data[0];
        f2 = f.data[1];
        VA_ASSERT(f2 - f1 > 1);
        ok = 2;

        for (f0 = f1 + 1; f0 <= f2 - 1; f0++) {
            for (i = 0; i < lmat.size_y; i++) {
                if (lmat.data[i * lmat.size_x + f0 - 1] != 4) {
                    ok--;
                    break;
                }
            }

            if (ok < 1)
                break;
        }

        if (ok >= 1) {
            if (pair_starts.size > 0) {
                pair_starts = va_reallocVecI32(pair_starts, pair_starts.size + 1);
                pair_starts.data[pair_starts.size - 1] = f1;
                pair_ends = va_reallocVecI32(pair_ends, pair_ends.size + 1);
                pair_ends.data[pair_ends.size - 1] = f2;
            }
            f.data = &f.data[2];
            f.size -= 2;
        } else {
            for (i = 0; i < ch_list->size_y; i++)
                ch_list->data[i * ch_list->size_x + f.data[0] - 1] = ALLELE_NONE;
            f.size--;
            f.data = &f.data[1];
        }

        /*
         * TODO: also consider 3 in a row color changes, where one is due to
         * gaps. This is partly done automatically since I'm including all
         * gaps, so for a 3-in-a-row, we'd have two pairs formed.
         */
    }


    fprintf(stderr, "\t\tGiven a list of %lu proposed SNPs\n", 
            (pair_starts.size > 0) ? pair_starts.size : 0);
    
    /* the outer loop, over the various thresholds */
    for (t = 0; t < thr_vec.size; t++){
        
        /* the inner loop, over the color disagreements found */
        for (i = 0; i < pair_starts.size; i++){
            int k;
            VA_UINT8 vd;

            /* if the color disagreements are consecutive */
            f1 = pair_starts.data[i];    /* the first position */
            f2 = pair_ends.data[i];      /* the second position */

            diColor1 = va_allocVecU32(f2 - f1 + 1); 
            diColor2 = va_allocVecU32(f2 - f1 + 1); 
            for (k = 0; k < diColor1.size; k++)
            {
                diColor1.data[k] = cmat.data[f1 + k - 1]; 
                diColor2.data[k] = cmat.data[cmat.size_x + f1 + k - 1]; 
            }

            letter1 = lmat.data[f1 - 1];
            letter2 = lmat.data[lmat.size_x + f1 - 1];

            /* fill in the ch_list with the appropriate type of action
             * (see destroy function), and return the support we have
             * for the top and the bottom allele (t and b, respectively)
             */
            vd = destroy(cdata, reads, readstarts, covmap,
                         f1, f2, diColor1, diColor2, letter1, letter2);


	    //fprintf(stderr, "destroystate: %i: vd: %u\n", f1, vd);

            for (k = 0;  k < pair_ends.data[i] - pair_starts.data[i] + 1; k++)
                ch_list->data[t * ch_list->size_x + pair_starts.data[i] - 1 + k] = vd; 
	    free(diColor1.data);
	    free(diColor2.data);
        }

        /* Quick steparound threshold. TODO: take it out if you reimplement
         * threshold.
         */
        for (i = 1; i < ch_list->size_y; i++)
            for (j = 1; j < ch_list->size_x; j++)
                ch_list->data[i * ch_list->size_x + j] = ch_list->data[j];
        break;
    }
    free(f_orig);
    if (pair_starts.data!=NULL) {
	free(pair_starts.data);
    }
    if (pair_ends.data!=NULL) {
	free(pair_ends.data);
    }
    rc = va_GetTime(&te);
    VA_ASSERT(rc == 0);

    tdiff = va_TimeSub(&te, &ts);
    fprintf(stderr, "\t\tDone destroy_state in: %ld.%03ld secs\n",
            tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);
}
    
/* function that decides whether to destroy or alter the SNP */
static VA_UINT8 destroy(
    CData_t     cdata,
    VA_Vector   reads,
    VA_VecI32   readstarts,
    Covmaps_t   covmap,
    VA_UINT32   f1, 
    VA_UINT32   f2,
    VA_VecU32   diColor1, 
    VA_VecU32   diColor2, 
    VA_UINT32   letter1, 
    VA_UINT32   letter2)
{
    VA_UINT8 type=ALLELE_NONE;
    int i;
    int j;
    int cTop = 0;
    int cBot = 0;
    VA_DOUBLE meancov;
    VA_DOUBLE cerr;
    VA_DOUBLE mincol; 
    VA_DOUBLE lerr;
    VA_DOUBLE loghet; 
    VA_DOUBLE loghomTop; 
    VA_DOUBLE loghomBot; 
    VA_DOUBLE letgapcor;
    VA_DOUBLE sum1;
    VA_DOUBLE sum2;
    VA_DOUBLE multfact;
    VA_DOUBLE lTop;
    VA_DOUBLE lBot;
    VA_DOUBLE lTopo;
    VA_DOUBLE lBoto;


    for (i = 0; i < diColor1.size; i++) {
        if (diColor1.data[i] == 5) {
		return ALLELE_NONE;
        }
    }
   
    for (i = 0; i < diColor2.size; i++) {
        if (diColor2.data[i] == 5) {
                return ALLELE_NONE;
        }
    }
   
        /* consider which reads are within constraints */
        VA_VecI32 goodreads = va_allocVecI32(readstarts.size);
        ReadsNode_t * rd = (ReadsNode_t *)reads.data;

        goodreads.size = 0;
        for (i = 0; i < readstarts.size; i++) {
            ReadsNode_t * read = &rd[i];
            if (readstarts.data[i] + (read->isLetterspace ? 0 : 1) <= f1) {
                //(readstarts.data[i] + 500 >= f2)) {
                goodreads.data[goodreads.size++] = i;
            }
        }
       
        if (goodreads.size == 0) {
            free(goodreads.data);
            goodreads.data = NULL;
        } else {
            goodreads = va_reallocVecI32(goodreads, goodreads.size);
        }

        for (i = 0; i < goodreads.size; i++) {
            VA_UINT32 r = goodreads.data[i];
            ReadsNode_t * read = &rd[r];
            
	    //if the read spans the region
            if (read->cs_seq.size > (f2-read->cs_chrstart)) { 
		int read_offset = f1 - read->cs_chrstart;
		VA_ASSERT(read_offset>=0);
		VA_ASSERT(diColor1.size==f2-f1+1);
		VA_ASSERT(read->cs_seq.size-read_offset>diColor1.size-1);
		int x=0;
		for (j=0; j<read->cs_seq.size; j++) {
			x+=read->cs_seq.data[j];
		}
		for (j=0; j<diColor1.size && (read->cs_seq.data[read_offset+j]==diColor1.data[j]); j++);
		if (j==diColor1.size) {
			cTop++;	
		}
		for (j=0; j<diColor2.size && (read->cs_seq.data[read_offset+j]==diColor2.data[j]); j++);
		if (j==diColor2.size) {
			cBot++;	
		}
            }    
            
        }        
	if (goodreads.size>0) {
		free(goodreads.data);
	}
        meancov = 0.0;
        for (j = 0; j < covmap.colors_positive.size_y; j++)
            meancov += covmap.colors_positive.data[j * covmap.colors_positive.size_x + f1 - 1] +
                   covmap.colors_positive.data[j * covmap.colors_positive.size_x + f2 - 1] +
		   covmap.colors_negative.data[j * covmap.colors_negative.size_x + f1 - 1] +
                   covmap.colors_negative.data[j * covmap.colors_negative.size_x + f2 - 1];
        meancov = meancov / 10;
            
        sum1 = covmap.colors_positive.data[diColor1.data[0] * covmap.colors_positive.size_x + f1 - 1] + 
               covmap.colors_positive.data[diColor2.data[0] * covmap.colors_positive.size_x + f1 - 1] +
	       covmap.colors_negative.data[diColor1.data[0] * covmap.colors_negative.size_x + f1 - 1] + 
               covmap.colors_negative.data[diColor2.data[0] * covmap.colors_negative.size_x + f1 - 1];
        sum2 = covmap.colors_positive.data[diColor1.data[1] * covmap.colors_positive.size_x + f2 - 1] + 
               covmap.colors_positive.data[diColor2.data[1] * covmap.colors_positive.size_x + f2 - 1] +
               covmap.colors_negative.data[diColor1.data[1] * covmap.colors_negative.size_x + f2 - 1] + 
               covmap.colors_negative.data[diColor2.data[1] * covmap.colors_negative.size_x + f2 - 1];
        mincol = (sum1 > sum2) ? sum2 : sum1;

        lTop = covmap.letters_positive.data[letter1 * covmap.letters_positive.size_x + f1 - 1] +
	       covmap.letters_negative.data[letter1 * covmap.letters_negative.size_x + f1 - 1];
        lBot = covmap.letters_positive.data[letter2 * covmap.letters_positive.size_x + f1 - 1] + 
	       covmap.letters_negative.data[letter2 * covmap.letters_negative.size_x + f1 - 1];

        if (((cTop + lTop < 2) && (cBot + lBot < 2)) ||
            (cTop + cBot < 0.7 * mincol) ||
            ((cTop + cBot) / 2.0 < meancov)) //||
	  //((cTop + cBot) > 10 && (cTop == 0 || cBot == 0)))
        {
            type = ALLELE_NONE;
            return (type);
        }
           
        /* probabilities... */
        
        /* METHOD 1: "0.5 model" */
        cerr = 0.04;
        //snp_pen = 0.001;
        lerr = 0.01;
        
        loghet = log(snp_rate) + (cTop + cBot) * log((1 - cerr)/2);
        loghomTop = log(1 - snp_rate) + (cTop) * log((1 - cerr)) + cBot * log(cerr);
        loghomBot = log(1 - snp_rate) + (cBot) * log((1 - cerr)) + cTop * log(cerr);
        
        lTopo = lTop;
        lBoto = lBot;
        if (lTop + lBot) {
            multfact = 1.0 * (cTop + cBot) / (lTop + lBot);
            if  (!isnan(multfact) && multfact > 1) {
                lTop = lTop * ((multfact - 1) / 2.0 + 1);
                lBot = lBot * ((multfact - 1) / 2.0 + 1);
            }
            
            loghet = loghet + (lTop + lBot) * log((1 - lerr)/2);
            loghomTop = loghomTop + (lTop) * log((1 - lerr)) + lBot * log(lerr);
            loghomBot = loghomBot + (lBot) * log((1 - lerr)) + lTop * log(lerr);
        
            letgapcor = 0.7 * 0.01; 
            if ((letter1 != letter2) && (letter1 == 4))
                loghet = loghet + lTop * log((letgapcor));
            else if ((letter1 != letter2) && (letter2 == 4))
                loghet = loghet + lBot * log((letgapcor));
        }
        
        if ((loghet >= loghomTop) && (loghet >= loghomBot))
            type = ALLELE_OK;
        else if (loghomTop > loghomBot)
            type = ALLELE_TOP;
        else if (loghomTop < loghomBot)
            type = ALLELE_BOT;
        else
        {
            int diC12 = FALSE;

            if (diColor1.size == diColor2.size)
            {
                diC12 = (memcmp(diColor1.data, diColor2.data, 
                                diColor1.size * sizeof(VA_UINT32)) == 0) ?  TRUE : FALSE;
            }

            if ((lTop + lBot == 0) && diC12)
            {
                type = ALLELE_OK;
            }
            else if ((lTop + lBot > 0) && diC12)
            {
                if (letter1 == letter2)
                    type = ALLELE_OK;
                else
                    /* it should be determnined by the het stuff... 
                     * This whould be a WEIRD CASE 
                    */
                    ; 
            }
            else
            {
                type = ALLELE_NONE;
            }
        } 
        
        /*
         * Method 3: ALTERNATIVE: DO SEPARATE VORES FOR HET OR HOM. IF BOTH VOTES FOR
         * ONE, its clear, otherwise... think about it
         */ 
        
        /* NEW CODE JAN 2010
         * WHY IS diColor1 EVER LONGER THAN 2?
         */
        {
            VA_VecU32 rc = va_allocVecU32(diColor1.size);
            int rcdi1 = FALSE;
            int rcdi2 = FALSE;
            int agree;
            VA_DOUBLE frac;
           
            for (i = 0; i < rc.size; i++)
                rc.data[i] = refcol.data[f1 - 1 + i];

            /* find which diColor the ref agrees with */
            if (diColor1.size == rc.size)
            {
                rcdi1 = (memcmp(diColor1.data, rc.data, 
                               diColor1.size * sizeof(VA_UINT32)) == 0) ?  TRUE : FALSE;
            }

            if (diColor2.size == rc.size)
            {
                rcdi2 = (memcmp(diColor2.data, rc.data, 
                               diColor2.size * sizeof(VA_UINT32)) == 0) ?  TRUE : FALSE;
            }

            free(rc.data);

            if (rcdi1)
            {
                agree = 0;
                frac = 1.0 * cTop / (cTop + cBot);
            }
            else if (rcdi2)
            {
                agree = 1;
                frac = 1.0 * cBot / (cTop + cBot);
            }
            else
            {
                agree = 2;
                frac = 0;
            }

            /* extra rule: if more than a third agree with ref, call het snp
             * kind of ignoring lspace here...
             */
            if (((cTop + cBot) < 15) && (frac >= 0.25) && ((lTop + lBot) < 3) && (frac <= 0.5))
            {
                type = ALLELE_OK;
            }
            else if (((cTop + cBot) < 15) && (frac >= 0.25) && ((lTopo + lBoto) < 3)) 
            {
                if (agree == 0)
                    type = ALLELE_TOP;
                else if (agree == 1)
                    type = ALLELE_BOT;
                else
                    type = ALLELE_NONE;
            }
            

            if ((lTop == 0) && (lBot == 0) && (cTop == 1) && (cBot == 1)) 
            {
                if (agree == 0)
                    type = ALLELE_TOP;
                else if (agree == 1)
                    type = ALLELE_BOT;
                else
                    type = ALLELE_NONE;
            }
 
            if (cTop + cBot + lTop + lBot < 2)
                type = ALLELE_NONE;
        }

    return (type);
}

